from ipycanvas import Canvas, hold_canvas, MultiCanvas
import ipywidgets as widgets
from itertools import product, chain, takewhile
import math
import random
import time
from lab_utils import *


class Connect4:
    def __init__(self, drop_disc_function, check_sequence_function, computer_move_function = None, layout_width = "40%"):
        self._drop_disc = drop_disc_function
        self._check_sequence = check_sequence_function
        self._computer_move = computer_move_function
        self._cell_size = 100
        self._disk_radius = int(round(self._cell_size * 0.75 / 2))        
        self.columns = 7
        self.rows = 6
        self.width = self._cell_size * self.columns
        self.height = self._cell_size * self.rows
        self._canvas = MultiCanvas(2, width = self.width, height = self.height)
        self._canvas.layout.height = "auto"
        
        self._buttons = [widgets.Button(description = "▼") for _ in range(self.columns)]
        for i, b in enumerate(self._buttons):
            b._column_index = i # Adds an attribute with its index
            b.on_click(self._on_button_clicked)
        
        hbox_buttons = widgets.HBox(self._buttons)
        self.output = widgets.Output()        
        
        self._draw_board()
        
        self._canvas_turn = Canvas(width = self._cell_size, height = self._cell_size)
        self._new_game_button = widgets.Button(description = "Start new game")
        self._new_game_button.on_click(self._start_new_game)
        
        main = widgets.VBox([hbox_buttons, self._canvas, self.output])
        main.layout.width = layout_width
        
        self._label = widgets.Label()
        self._label.layout.align_self = "center"
        controls = widgets.VBox([widgets.HBox([self._label, self._canvas_turn]), self._new_game_button])
        controls.layout.margin = "20px"

        self._start_game()        
        
        display(widgets.HBox([main, controls]))

        
    def _update_turn(self):
        c = self._canvas_turn
        with hold_canvas(c):
            c.clear()
            self._draw_disc(0, 0, self.turn, c)
        
        
    def _start_new_game(self, b):
        with self.output:
            self._start_game()
    
    
    def random_choice(self, values):
        return random.choice(values)
    
        
    def _start_game(self):
        self.cells = [[0] * self.columns for _ in range(self.rows)]   
        self.turn = 1
        self._canvas[0].clear()
        self._enable_buttons()
        self._label.value = "Current turn:"
        self._update_turn()

        
    def _change_turn(self):
        self.turn = 1 if self.turn == 2 else 2
        self._update_turn()
        
        
    def _on_button_clicked(self, b):
        with self.output:
            if not b.disabled:
                self._drop(b._column_index)
            
            
    def _drop(self, column):
        row = self._drop_disc(self, column)
        if row >= 0:
            self._move(column, row)
            
            
    def _finish_game(self, text):
        self._label.value = text
        for x in range(self.columns):
            self._buttons[x].disabled = True            

            
    def _move(self, column, row):
        steps = 20
        c = self._canvas[0]
        prev_y = None
        rs = row*steps
        for y in chain(range(-steps, rs, 4), range(rs, rs-2, -1), range(rs-2, rs)):
            with hold_canvas(c):        
                if prev_y is not None:
                    c.clear_rect(column * self._cell_size, prev_y / steps * self._cell_size, self._cell_size, self._cell_size)
                self._draw_disc(column, y / steps, self.turn, c)
            time.sleep(0.03)
            prev_y = y
                
        self.cells[row][column] = self.turn
        self._draw_disc(column, row, self.turn, c)
        
        if self._check_sequence(self, column, row, 4):            
            self._finish_game("WINNER:")
        elif all(v!=0 for v in self.cells[0]):
            self._finish_game("DRAW")
        else:
            self._change_turn()
            self._enable_buttons()
            if self.turn == 2 and self._computer_move:
                computer_column = self._computer_move(self)
                if computer_column < 0 or computer_column >= self.columns or self.cells[0][computer_column] != 0:
                    raise Exception(f"Invalid computer move (column {computer_column})!")
                self._drop(computer_column)
            
            
    def _enable_buttons(self):
        for x in range(self.columns):
            self._buttons[x].disabled = self.cells[0][x] != 0
        
        
    def _draw_board(self):
        cs2 = self._cell_size // 2
        c = self._canvas[1]
        c.fill_style = "blue"
        c.fill_rect(0, 0, self.width, self.height)
        c.begin_path()            
        for y, x in product(range(self.rows), range(self.columns)):
            cx, cy = x * self._cell_size + cs2, y * self._cell_size + cs2
            c.move_to(cx, cy)
            c.arc(cx, cy, self._disk_radius, 0, math.pi * 2, True)
        c.clip()
        c.close_path()
        c.clear()
        
        
    def _draw_disc(self, x, y, player, canvas):
        canvas.fill_style = "red" if player == 1 else "yellow"
        canvas.fill_circle(x * self._cell_size + self._cell_size // 2, y * self._cell_size + self._cell_size // 2, self._disk_radius)

        
        
        
def _fake_check_sequence(game, x, y, length):
    return False

def _get_functions(g, drop_disc_function_name, check_sequence_function_name, computer_move_function_name):
    drop_disc_function = g.get(drop_disc_function_name)
    if not callable(drop_disc_function):
        print(f"Function '{drop_disc_function_name}' not defined.")
        return None
    
    check_sequence_function = g.get(check_sequence_function_name) if check_sequence_function_name else _fake_check_sequence
    if not callable(check_sequence_function):
        print(f"Function '{check_sequence_function_name}' not defined.")
        return None

    computer_move_function = g.get(computer_move_function_name) if computer_move_function_name else None
    if check_sequence_function and not callable(check_sequence_function):
        print(f"Funcion '{computer_move_function_name}' not defined.")
        return None
        
    return drop_disc_function, check_sequence_function, computer_move_function
    
    
c1, c2, c3 = None, None, None    
    
def es_1(g):
    funcs = _get_functions(g, "drop_disc", None, None)
    if not funcs:
        return
    c1 = Connect4(*funcs)


def es_2(g):
    funcs = _get_functions(g, "drop_disc", "check_sequence", None)
    if not funcs:
        return
    c2 = Connect4(*funcs)  

    
def es_3(g):
    funcs = _get_functions(g, "drop_disc", "check_sequence", "computer_move")
    if not funcs:
        return
    c3 = Connect4(*funcs)
    
